#region Copyright Syncfusion Inc. 2001-2016.
// Copyright Syncfusion Inc. 2001-2016. All rights reserved.
// Use of this code is subject to the terms of our license.
// A copy of the current license can be obtained at any time by e-mailing
// licensing@syncfusion.com. Any infringement will be prosecuted under
// applicable laws. 
#endregion
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using System;
using System.IO;
using System.Net;
using System.Net.Security;
using System.Security.Cryptography.X509Certificates;
using System.Text.RegularExpressions;
using System.Timers;

namespace WebHDFS
{
    class Program
    {
        static string userName, password,domainName, hostName, exitOption;
        static Timer checkClusterState;
        static int clusterOption;
        static string primaryNamNode, StandbyNameNode;
        static void Main(string[] args)
        {
            try
            {
                bool clusterChoice = true;
                var clusterType = ClusterType.Normal;
                checkClusterState = new Timer();
                checkClusterState.Interval = 10000;
                checkClusterState.Elapsed += checkClusterState_Elapsed;
                var commonHome = Directory.GetParent("..\\..\\..\\..\\..\\..\\..\\..\\").ToString();
                do
                {
                    Console.Write("Enter Primary Namenode IP : ");
                    primaryNamNode = Console.ReadLine();
                    Console.Write("Enter Secondary Namenode IP : ");
                    StandbyNameNode = Console.ReadLine();
                    hostName = primaryNamNode;
                    Console.WriteLine("Please choose the cluster type (For localhost please choose normal cluster)\n1.Normal Cluster\n2.Secure Cluster\n3.Azure Cluster");
                    clusterOption = int.Parse(Console.ReadLine());
                    if (clusterOption >= 1 && clusterOption <= 3)
                    {
                        switch (clusterOption)
                        {
                            case 1:
                                clusterType = ClusterType.Normal;
                                if (hostName != "localhost" || hostName != "127.0.0.1")
                                    checkClusterState.Enabled = true;
                                break;
                            case 2:
                                clusterType = ClusterType.Secure;
                                Console.WriteLine(
                                    "Please enter the UserName, Password and Domain Name\nUserName:");
                                userName = Console.ReadLine();
                                Console.WriteLine("Password:");
                                password = Console.ReadLine();
                                Console.WriteLine("Domain Name:");
                                domainName = Console.ReadLine();
                                ServicePointManager.ServerCertificateValidationCallback += ValidateServerCertificate;
                                hostName = GetHostName(hostName);
                                checkClusterState.Enabled = true;
                                break;
                            case 3:
                                clusterType = ClusterType.Azure;
                                Console.WriteLine("Please enter the UserName and Password\nUserName:");
                                userName = Console.ReadLine();
                                Console.WriteLine("password:");
                                password = Console.ReadLine();
                                checkClusterState.Enabled = true;
                                break;
                        }
                        int i;
                        do
                        {
                            Console.WriteLine("Enter any operation");
                            Console.WriteLine("1. Upload");
                            Console.WriteLine("2. Delete");
                            Console.WriteLine("3. MoveFile");
                            Console.WriteLine("4. MakeDirectory");
                            Console.WriteLine("5. SetReplication");
                            Console.WriteLine("6. Exit");
                            i = int.Parse(Console.ReadLine());
                            var p = new Program();
                            switch (i)
                            {
                                case 1:
                                    Console.WriteLine("Uploading file from '" + commonHome +
                                                      "\\BigDataSDK\\Samples\\Data\\WarPeace.txt'" +
                                                      " to hdfs location " + "'/Data/PracticeTemp1.txt'");
                                    p.UploadFile(hostName, clusterType,
                                        commonHome + "\\BigDataSDK\\Samples\\Data\\WarPeace.txt",
                                        "/Data/PracticeTemp.txt");
                                    break;
                                case 2:
                                    Console.WriteLine("Deleting file from hdfs location " + "'/Data/PracticeTemp.txt'");
                                    if (p.DeleteFile(hostName, clusterType, "/Data/PracticeTemp.txt"))
                                        Console.WriteLine("Delete operation completed");
                                    else
                                        Console.WriteLine("Delete operation Failed");
                                    break;
                                case 3:
                                    Console.WriteLine("Move file from hdfs location " +
                                                      "'/Data/PracticeTemp.txt' to hdfs location " +
                                                      "'/Data/PracticeTemp2.txt'");
                                    if (p.MoveFile(hostName, clusterType, "/Data/PracticeTemp.txt",
                                        "/Data/PracticeTemp2.txt"))
                                        Console.WriteLine("Move operation completed");
                                    else
                                        Console.WriteLine("Move operation Failed");
                                    break;
                                case 4:
                                    Console.WriteLine("Make directory from hdfs location " + "'/Data/PracticeDirectory'");
                                    if (p.CreateDirectory(hostName, clusterType, "/Data/PracticeDirectory"))
                                        Console.WriteLine("Make directory operation completed");
                                    else
                                        Console.WriteLine("Make directory operation Failed");
                                    break;
                                case 5:
                                    Console.WriteLine("set replication 5 for hdfs location " +
                                                      "'/Data/PracticeTemp.txt'");
                                    if (p.SetReplication(hostName, clusterType, "/Data/PracticeTemp.txt", 5))
                                        Console.WriteLine("Set replication operation completed");
                                    else
                                        Console.WriteLine("Set Replication operation Failed");
                                    break;
                                case 6:
                                    break;
                            }
                        } while (i != 6);
                        if (i == 6)
                        {
                            Console.WriteLine(
                                "Press 'Y' to continue by connecting new cluster or Press any key to exit:");
                            exitOption = Console.ReadLine();
                            if (exitOption.ToUpper() != "Y")
                                break;
                        }
                    }
                    else
                    {
                        Console.WriteLine("Invalid Choice\nPress 'Y' to try again or Press any key to exit:");
                        clusterChoice = (Console.ReadLine().ToUpper() == "Y");
                    }
                } while (clusterChoice);
            }
            catch (Exception ex)
            {
                //TODO
            }
        }

        /// <summary>
        /// Upload file operation
        /// </summary>
        /// <param name="host"></param>
        /// <param name="clusterType"></param>
        /// <param name="fileToUpload"></param>
        /// <param name="hdfspathToUpload"></param>
        private void UploadFile(string host, ClusterType clusterType, string fileToUpload, string hdfspathToUpload)
        {
            try
            {
                string url = string.Empty;
                hdfspathToUpload = hdfspathToUpload.TrimEnd('/').TrimStart('/');

                //Creating the url
                url = (clusterType.Equals(ClusterType.Secure))
                    ? ("https://" + host + ":50470/webhdfs/v1/" + hdfspathToUpload + "/?op=CREATE&data=true").Replace(
                        "#", "%23")
                    : (clusterType.Equals(ClusterType.Azure))
                        ? ("https://" + host + ":8000/webhdfs/v1/" + hdfspathToUpload + "/?user.name=" + userName +
                           "&op=CREATE&data=true").Replace("#", "%23").Replace("+", "%2b")
                        : ("http://" + host + ":50070/webhdfs/v1/" + hdfspathToUpload +
                           "/?user.name=SYSTEM&op=CREATE&data=true").Replace("#", "%23");

                var webRequest = WebRequest.Create(url);
                if (clusterType.Equals(ClusterType.Secure))
                    webRequest.Credentials = new NetworkCredential(userName, password);
                else if (clusterType.Equals(ClusterType.Azure))
                {
                    webRequest.Headers.Add("Username", userName);
                    webRequest.Headers.Add("Password", password);
                }
                webRequest.Method = "PUT";
                const int chunkSize = 1024 * 1024; // read the file by chunks of 10 MB
                using (var file = new FileStream(fileToUpload, FileMode.Open, FileAccess.Read))
                {
                    var buffer = new byte[file.Length];
                    long fileLength = file.Length;
                    if (fileLength <= chunkSize)
                        buffer = new byte[file.Length];
                    webRequest.ContentType = "application/octet-stream";
                    webRequest.ContentLength = buffer.Length;
                    file.Read(buffer, 0, buffer.Length);
                    Stream objStream = webRequest.GetRequestStream();
                    objStream.Write(buffer, 0, buffer.Length);

                    objStream.Close();
                    objStream = webRequest.GetResponse().GetResponseStream();
                    url = (clusterType.Equals(ClusterType.Secure))
                        ? ("https://" + host + ":50470/webhdfs/v1/" + hdfspathToUpload +
                           "/?op=SETREPLICATION&replication=1").Replace("#", "%23")
                        : (clusterType.Equals(ClusterType.Azure))
                            ? ("https://" + host + ":8000/webhdfs/v1/" + hdfspathToUpload + "/?user.name=SYSTEM" +
                               "&op=SETREPLICATION&replication=1").Replace("#", "%23")
                            : ("http://" + host + ":50070/webhdfs/v1/" + hdfspathToUpload +
                               "/?user.name=SYSTEM" + "&op=SETREPLICATION&replication=1").Replace("#", "%23");
                    SendWebRequest(url, "PUT", clusterType);

                    buffer = (file.Length - file.Position) > chunkSize ? new byte[chunkSize] : new byte[(file.Length - file.Position)];
                    while (file.Read(buffer, 0, buffer.Length) > 0)
                    {
                        //Append File
                        url = (clusterType.Equals(ClusterType.Secure))
                            ? ("https://" + host + ":50470/webhdfs/v1/" +
                               hdfspathToUpload +
                               "/?Op=APPEND&buffersize=" + webRequest.ContentLength)
                            : (clusterType.Equals(ClusterType.Azure))
                                ? ("https://" + host + ":8000/webhdfs/v1/" + hdfspathToUpload + "/?user.name=" +
                                   userName + "&op=APPEND&data=true")
                                : ("http://" + host + ":50070/webhdfs/v1/" +
                                   hdfspathToUpload +
                                   "/?user.name=SYSTEM&Op=APPEND&buffersize=" + webRequest.ContentLength);
                        WebRequest webRequest1 = WebRequest.Create(url);
                        webRequest1.Credentials = new NetworkCredential(userName, password);
                        webRequest1.Method = "POST";
                        webRequest1.ContentType = "application/octet-stream";
                        webRequest1.ContentLength = buffer.Length;
                        Stream objStreamAppend = webRequest1.GetRequestStream();
                        objStreamAppend.Write(buffer, 0, buffer.Length);
                        objStreamAppend.Close();
                        webRequest1.GetResponse().GetResponseStream();
                        buffer = (file.Length - file.Position) > chunkSize ? new byte[chunkSize] : new byte[(file.Length - file.Position)];
                    }
                }
                if (host == "localhost" || host == "127.0.0.1")
                    url = ("http://" + host + ":50070/webhdfs/v1/" + hdfspathToUpload +
                         "/?user.name=SYSTEM" + "&op=SETREPLICATION&replication=1").Replace("#", "%23");
                else
                {
                    url = (clusterType.Equals(ClusterType.Secure))
                        ? ("https://" + host + ":50470/webhdfs/v1/" + hdfspathToUpload +
                           "/?op=SETREPLICATION&replication=3").Replace("#", "%23")
                        : (clusterType.Equals(ClusterType.Azure))
                            ? ("https://" + host + ":8000/webhdfs/v1/" + hdfspathToUpload + "/?user.name=SYSTEM" +
                               "&op=SETREPLICATION&replication=3").Replace("#", "%23")
                            : ("http://" + host + ":50070/webhdfs/v1/" + hdfspathToUpload +
                               "/?user.name=SYSTEM" + "&op=SETREPLICATION&replication=3").Replace("#", "%23");
                }
                SendWebRequest(url, "PUT", clusterType);
            }
            catch (Exception e)
            {
                Console.Error.WriteLine(
                    "/**UploadHdfsFile(string nameNodeIP, string fileToUpload, string hdfspathToUpload)**/" +
                    e.Message + "\n" + e.StackTrace);
            }
        }

        /// <summary>
        /// Deletes a selected file
        /// </summary>
        /// <param name="clusterType"></param>
        /// <param name="filePath">HDFS path of file needs to be deleted</param>
        /// <param name="host"></param>
        /// <returns>If Delete succeeds returns true else returns false</returns>
        public bool DeleteFile(string host, ClusterType clusterType, string filePath)
        {
            string url = (clusterType.Equals(ClusterType.Secure))
                ? ("https://" + host + ":50470/webhdfs/v1/" + filePath.TrimStart('/') + "/?op=DELETE&recursive=true")
                    .Replace("#", "%23")
                : (clusterType.Equals(ClusterType.Azure))
                    ? ("https://" + host + ":8000/webhdfs/v1/" + filePath.TrimStart('/') + "/?user.name=" + userName +
                       "&op=DELETE&recursive=true").Replace("#", "%23").Replace("+", "%2b")
                    : ("http://" + host + ":50070/webhdfs/v1/" + filePath.TrimStart('/') +
                       "/?user.name=SYSTEM&op=DELETE&recursive=true").Replace("#", "%23").Replace("+", "%2b");
            return SendWebRequest(url, "DELETE", clusterType);
        }

        /// <summary>
        /// Creates a new folder
        /// </summary>
        /// <param name="clusterType"></param>
        /// <param name="folderPath">HDFS Path of folder to be created</param>
        /// <param name="host"></param>
        /// <returns>If folder is created returns true else returns false</returns>
        public bool CreateDirectory(string host, ClusterType clusterType, string folderPath)
        {
            string url = (clusterType.Equals(ClusterType.Secure))
                ? ("https://" + host + ":50470/webhdfs/v1/" + folderPath.TrimStart('/') + "/?op=MKDIRS").Replace("#",
                    "%23")
                : (clusterType.Equals(ClusterType.Azure))
                    ? ("https://" + host + ":8000/webhdfs/v1/" + folderPath.TrimStart('/') + "/?user.name=" + userName +
                       "&op=MKDIRS").Replace("#", "%23").Replace("+", "%2b")
                    : ("http://" + host + ":50070/webhdfs/v1/" + folderPath.TrimStart('/') +
                       "/?user.name=SYSTEM&op=MKDIRS")
                        .Replace("#", "%23");
            return SendWebRequest(url, "PUT", clusterType);
        }

        /// <summary>
        ///  Renames Selected file/folder(s)
        /// </summary>
        /// <param name="host"></param>
        /// <param name="clusterType"></param>
        /// <param name="sourceName">Current file name</param>
        /// <param name="targetName">New name for the file</param>
        /// <returns>If Rename succeeds returns true else returns false</returns>
        public bool MoveFile(string host, ClusterType clusterType, string sourceName, string targetName)
        {
            string commonHome = Directory.GetParent("..\\..\\..\\..\\..\\..\\..\\..\\").ToString();
            UploadFile(host, clusterType, commonHome + "\\BigDataSDK\\Samples\\Data\\WarPeace.txt", targetName);
            DeleteFile(host, clusterType, "/Data/PracticeTemp.txt");
            return true;
        }

        /// <summary>
        ///  Renames Selected file/folder(s)
        /// </summary>
        /// <param name="host"></param>
        /// <param name="clusterType"></param>
        /// <param name="sourceName">Current file name</param>
        /// <param name="replication"></param>
        /// <returns>If Rename succeeds returns true else returns false</returns>
        public bool SetReplication(string host, ClusterType clusterType, string sourceName, int replication)
        {
            string url = (clusterType.Equals(ClusterType.Secure))
                ? ("https://" + host + ":50470/webhdfs/v1" + sourceName + "/?op=SETREPLICATION&replication=" +
                   replication).Replace("#", "%23")
                : (clusterType.Equals(ClusterType.Azure))
                    ? ("https://" + host + ":8000/webhdfs/v1/" + sourceName + "/?user.name=SYSTEM" +
                       "&op=SETREPLICATION&replication=" + replication).Replace("#", "%23")
                    : ("http://" + host + ":50070/webhdfs/v1" + sourceName +
                       "/?user.name=SYSTEM&op=SETREPLICATION&replication=" + replication).Replace("#", "%23");
            return SendWebRequest(url, "PUT", clusterType);
        }

        /// <summary>
        /// Send webrequest for given URL and protocol
        /// </summary>
        /// <param name="url">URL</param>
        /// <param name="protocol">Protocol</param>
        /// <param name="clusterType"></param>
        /// <returns>Returns success status of web request <true>return true if success</true></returns>
        private bool SendWebRequest(string url, string protocol, ClusterType clusterType)
        {
            try
            {
                WebRequest request = WebRequest.Create(url);
                if (clusterType.Equals(ClusterType.Secure))
                    request.Credentials = new NetworkCredential(userName, password);
                else if (clusterType.Equals(ClusterType.Azure))
                {
                    request.Headers.Add("Username", userName);
                    request.Headers.Add("Password", password);
                }
                request.Method = protocol;
                using (var response = request.GetResponse())
                {
                    response.Close();
                }
                return true;
            }
            catch (Exception e)
            {
                Console.Error.WriteLine("/**SendWebRequest(string url, string protocol)**/" + e.Message + "\n" + e.StackTrace);
                return false;
            }
        }

        /// <summary>
        /// Certificate validation
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="certificate"></param>
        /// <param name="chain"></param>
        /// <param name="sslPolicyErrors"></param>
        /// <returns></returns>
        public static bool ValidateServerCertificate(object sender, X509Certificate certificate,
            X509Chain chain,
            SslPolicyErrors sslPolicyErrors)
        {
            return true;
        }

        /// <summary>
        /// Used to get host name
        /// </summary>
        /// <param name="ip"></param>
        /// <returns></returns>
        public static string GetHostName(string ip)
        {
            try
            {
                var hostEntry = Dns.GetHostByAddress(ip);
                return hostEntry.HostName;
            }
            catch (Exception ex)
            {
                return ip;
            }
        }

        #region Namenode Status

        /// <summary>
        /// Triggerd when check cluster status timer getting enabled in every 10seconds.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        static void checkClusterState_Elapsed(object sender, ElapsedEventArgs e)
        {
            if (IsNameNodeActive(hostName) != "active")
            {
                if (IsNameNodeActive(StandbyNameNode) == "active")
                {
                    hostName = GetHostName(StandbyNameNode);
                    StandbyNameNode = primaryNamNode;
                    primaryNamNode = hostName;
                }
            }
        }

        /// <summary>
        /// Check the status of the namenode
        /// </summary>
        /// <param name="hostName">It defines the host name of the connected cluster</param>
        /// <returns>returns the namonode status.</returns>
        public static string IsNameNodeActive(string hostName)
        {
            string nodeStatus = "";
            string nameSystem = GetNameSystemResponse(hostName);
            if (!string.IsNullOrEmpty(nameSystem))
            {
                JToken status = null;
                dynamic results = JsonConvert.DeserializeObject(nameSystem);
                JArray clusterDetailsArray = results["beans"];
                foreach (var jToken in clusterDetailsArray)
                {
                    var root = (JObject)jToken;
                    root.TryGetValue("tag.HAState", out status);
                }
                nodeStatus = status.ToString();
            }
            return nodeStatus; ;
        }

        #endregion Namenode Status.

        #region Get Response methods

        /// <summary>
        /// Set the Url based on the cluster and returns the response
        /// </summary>
        public static string GetNameSystemResponse(string hostName)
        {
            string namenodeStatus = string.Empty;
            if (clusterOption == 1)
                namenodeStatus = GetResponse("http://" + hostName + ":50070/jmx?qry=Hadoop:service=NameNode,name=FSNamesystem");
            else if (clusterOption == 2)
                namenodeStatus = GetSecuredWebRequestAndResponse("https://" + hostName + ":50470/jmx?qry=Hadoop:service=NameNode,name=FSNamesystem");
            else
                namenodeStatus = GetResponse("https://" + hostName + ":8000/jmx?qry=Hadoop:service=NameNode,name=FSNamesystem");
            return namenodeStatus;
        }

        #endregion

        #region To get Response for the url

        /// <summary>
        /// Gets response for given url
        /// </summary>
        /// <param name="url"></param>
        /// <returns>Returns data from URL</returns>
        public static string GetResponse(string url)
        {
            WebRequest request = WebRequest.Create(url);
            request.Method = "GET";
            using (Stream objStream = request.GetResponse().GetResponseStream())
            {
                using (StreamReader objReader = new StreamReader(objStream))
                {
                    return objReader.ReadToEnd();
                }
            }
        }

        /// <summary>
        /// Gets response for given secured url
        /// </summary>
        /// <param name="url"></param>
        /// <returns>Returns data from URL</returns>
        public static string GetSecuredWebRequestAndResponse(string sURL)
        {
            string error = string.Empty;
            HttpWebRequest connectionReq = null;
            userName = GetUsernameWithDomain(userName, domainName);
            try
            {
                if (!string.IsNullOrEmpty(sURL))
                {
                    connectionReq = (HttpWebRequest)WebRequest.Create(sURL);
                    connectionReq.Credentials = new NetworkCredential(userName, password);
                    connectionReq.Timeout = 100000;
                    var httpResponse = (HttpWebResponse)connectionReq.GetResponse();
                    var dataStream = httpResponse.GetResponseStream();
                    var reader = new StreamReader(dataStream);
                    var responseFromServer = reader.ReadToEnd();
                    return responseFromServer;
                }
                return string.Empty;
            }
            catch (WebException wex)
            {
                var Pagecontent = string.Empty;
                var response = (HttpWebResponse)wex.Response;
                if (wex.Response != null)
                {
                    Pagecontent = new StreamReader(wex.Response.GetResponseStream()).ReadToEnd();
                }
                try
                {
                    if (response.StatusCode == HttpStatusCode.Unauthorized)
                    {
                        try
                        {
                            connectionReq = (HttpWebRequest)WebRequest.Create(wex.Response.ResponseUri.AbsoluteUri);
                            connectionReq.Credentials = new NetworkCredential(userName, password, domainName);
                            connectionReq.Timeout = 100000;
                            var httpResponse = (HttpWebResponse)connectionReq.GetResponse();
                            var dataStream = httpResponse.GetResponseStream();
                            var reader = new StreamReader(dataStream);
                            var responseFromServer = reader.ReadToEnd();
                            return responseFromServer;
                        }
                        catch (WebException internalWebEx)
                        {

                            if (internalWebEx.Response != null)
                            {
                                Pagecontent = new StreamReader(internalWebEx.Response.GetResponseStream()).ReadToEnd();
                            }
                        }
                    }
                }
                catch (Exception ex)
                {
                }
                return "Fail; " + Pagecontent;
            }
            catch (Exception e)
            {
                return "Fail; " + e;
            }
        }

        /// <summary>
        /// Get the Username of the cluster with domain name.
        /// </summary>
        /// <param name="username">It defines the username of the cluster.</param>
        /// <param name="domain">It defines the domain name of the cluster name.</param>
        /// <returns>It returns the username with domain name.</returns>
        public static string GetUsernameWithDomain(string username, string domain)
        {
            if (string.IsNullOrEmpty(domain))
                return username;
            if (username.Contains("@") && username.Contains(domain))
                return username;
            return username + "@" + domain;
        }

        #endregion

        public enum ClusterType
        {
            Normal,
            Secure,
            Azure
        }
    }
}
